/*
  This file is part of Solver.

    Solver is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    Solver is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Solver. If not, see <http://www.gnu.org/licenses/>.
*/

package infra;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.jacop.constraints.Not;
import org.jacop.constraints.SumInt;
import org.jacop.constraints.XeqC;
import org.jacop.constraints.XplusYgtC;
import org.jacop.constraints.XplusYplusQgtC;
import org.jacop.core.Domain;
import org.jacop.core.IntVar;
import org.jacop.core.Store;
import org.jacop.search.DepthFirstSearch;
import org.jacop.search.IndomainRandom;
import org.jacop.search.RandomSelect;
import org.jacop.search.Search;
import org.jacop.search.SelectChoicePoint;

import infra.definition.CalendarWeek;
import infra.definition.Employee;
import infra.definition.LeaveOfAbsence;
import infra.definition.ProblemDefinition;
import infra.definition.ProblemXMLLoader;
import infra.definition.ScheduleRequest;
import infra.definition.TimePeriod;
import infra.definition.Watch;
import infra.formulation.CSPVariable;
import infra.formulation.ProblemFormulation;
import infra.solution.ProblemSolution;
import infra.solver.SolverConfig;

public class Solver {
	
	private SolverConfig solverConfig;
	
	public Solver(SolverConfig solverConfig) {
		this.solverConfig = solverConfig;
	}
	
	public ProblemSolution solve(String problemFile, String existingSolutionFile, String currentDate) {
		System.out.println("SILNIK WYSZUKIWANIA, WERSJA 7");
		ProblemDefinition definition = new ProblemXMLLoader().load(problemFile);
		definition.print();
		
		if (existingSolutionFile != null) {
			definition.existingSolution = new ProblemSolution(definition);
			definition.existingSolution.buildFromFile(existingSolutionFile);
			definition.recomputationDate = Utils.parse(currentDate);
			System.out.println("ISTNIEJACE ROZWIAZANIE");
			definition.existingSolution.printout();
			System.out.println("ZMIANA OD: " + Utils.build(definition.recomputationDate));
		}
				
		ProblemFormulation formulation = cspProblemFormulation(definition);
		if (formulation == null) {
			System.out.println("Bledna definicja problemu. Przerywam.");
			return null;
		}
        
		ProblemSolution solution = search(definition, formulation);
        if (solution != null) {
        	//ProblemSolution pSolution = new ProblemSolution(definition);
        	//pSolution.buildFromSolver6(solution);
        	solution.printout();        	
        }
        return solution;
	}
	
	private boolean printOnce = true;
	
	private IntVar buildVariable(ProblemDefinition definition, List<List<Integer>> employeesOffDays, List<List<Integer>> employeesWorkDays, Store store, int i, int j, String varType) {
		Calendar iDate = definition.getAnalysisPeriod().allDays.get(i);
		
		String wName = varType + "_" + (i + 1);			
		String pName = "EMP_" + (j + 1);
		String name = pName + "_" + wName;
			
		List<Integer> offDays = employeesOffDays.get(j);
		List<Integer> workDays = employeesWorkDays.get(j);
		int minimum = 0;
		int maximum = 1;
		if (offDays.contains(i)) {
			minimum = 0;
			maximum = 0;
		}
		if (workDays.contains(i)) {
			minimum = 1;
			maximum = 1;
		}
			
		if (definition.existingSolution != null) {
			if (definition.recomputationDate.after(iDate)) {
				String code = definition.existingSolution.data[j][i];
				if ((code == null) || (!code.equals(varType))) {
					minimum = 0;
					maximum = 0;
				} else {
					minimum = 1;
					maximum = 1;
				}
			}
		}
			
		IntVar personWatch = new IntVar(store, name, minimum, maximum);
		return personWatch;
	}
	
	private ProblemFormulation cspProblemFormulation(ProblemDefinition definition) {
		//PROBLEM DEFINITION
		Store store = new Store();
		
		TimePeriod calendar = definition.getAnalysisPeriod();
		List<Employee> employees = definition.getEmployees();
		Map<String, Watch> daysDefinition = definition.getDaysDefinition();
						
		int requiredNightWatches = 0;
		
		int totalMinFullWatches = 0;
		int totalMaxFullWatches = 0;
		int totalMinFullDayWatches = 0;
		int totalMaxFullDayWatches = 0;
		int totalMinFullNightWatches = 0;
		int totalMaxFullNightWatches = 0;
		int totalMinPartWatches = 0;
		int totalMaxPartWatches = 0;
		
		for (int i = 0; i < calendar.getAllDaysPerPeriod(); i++) {
			Watch dayDef = daysDefinition.get(calendar.getDayCode(i));
		    
		    totalMinFullWatches = totalMinFullWatches + dayDef.minDayWatch + dayDef.minNightWatch;
		    totalMaxFullWatches = totalMaxFullWatches + dayDef.maxDayWatch + dayDef.maxNightWatch;
		    
		    totalMinFullDayWatches = totalMinFullDayWatches + dayDef.minDayWatch;
		    totalMaxFullDayWatches = totalMaxFullDayWatches + dayDef.maxDayWatch;
		    
		    totalMinFullNightWatches = totalMinFullNightWatches + dayDef.minNightWatch;
		    totalMaxFullNightWatches = totalMaxFullNightWatches + dayDef.maxNightWatch;
		    
		    totalMinPartWatches = totalMinPartWatches + dayDef.minPartWatch;
		    totalMaxPartWatches = totalMaxPartWatches + dayDef.maxPartWatch;
		    
		    requiredNightWatches = requiredNightWatches + dayDef.minNightWatch;
		    
		    if (printOnce) {
		    	System.out.println("Dzien " + (i + 1) + " (" + calendar.getDayOfWeek(i) + ") : typ = " + dayDef.code);
		    }
		}
		
		if (printOnce) {
			System.out.println("Wymogi obsady okresu");
			
			System.out.println("Minimalna liczba pelnych wacht (dziennych i nocnych) : " + totalMinFullWatches);
			System.out.println("Maksymalna liczba pelnych wacht (dziennych i nocnych): " + totalMaxFullWatches);
			
			System.out.println("Minimalna liczba pelnych wacht - dziennych           : " + totalMinFullDayWatches);
			System.out.println("Maksymalna liczba pelnych wacht - dziennych          : " + totalMaxFullDayWatches);
			
			System.out.println("Minimalna liczba pelnych wacht - nocnych             : " + totalMinFullNightWatches);
			System.out.println("Maksymalna liczba pelnych wacht - nocnych            : " + totalMaxFullNightWatches);
			
			System.out.println("Minimalna liczba wacht resztkowych                   : " + totalMinPartWatches);
			System.out.println("Maksymalna liczba wacht resztkowych                  : " + totalMaxPartWatches);
		}
		
		//double minDailyRatio = totalMinFullDayWatches / (double)totalMinFullWatches;
		//double maxDailyRatio = totalMaxFullDayWatches / (double)totalMaxFullWatches;
				
		int[] minsPerPeriod = new int[employees.size()];
		int[] watchesPerPeriod = new int[employees.size()];
		int[] partWatchesPerPeriod = new int[employees.size()];
		
		for (int i = 0; i < employees.size(); i++) {
			minsPerPeriod[i] = definition.employeeMinsPerPeriod(i);
			watchesPerPeriod[i] = minsPerPeriod[i] / definition.getMinsPerWatch();
			int remainingMins = minsPerPeriod[i] % definition.getMinsPerWatch();
			if ((remainingMins > 0) && (remainingMins < definition.getMinimumShortWatchMins()) && (watchesPerPeriod[i] > 0)) {
				watchesPerPeriod[i]--;
				remainingMins = remainingMins + definition.getMinsPerWatch();
			}
			partWatchesPerPeriod[i] = remainingMins > 0 ? (remainingMins / definition.getMinsPerWatch()) + 1 : 0;
		}
						
		//ALL VARIABLE LISTS
		List<IntVar> allAllocations = new ArrayList<>();
		
		List<List<IntVar>> perDayAllocations = new ArrayList<>();
		List<List<IntVar>> perNightAllocations = new ArrayList<>();
		List<List<IntVar>> perPartAllocations = new ArrayList<>();
				
		List<List<IntVar>> perPersonDayAllocations = new ArrayList<>();
		List<List<IntVar>> perPersonNightAllocations = new ArrayList<>();
		List<List<IntVar>> perPersonPartAllocations = new ArrayList<>();
		
		int totalEmployeeFullWatches = 0;
		int totalEmployeePartWatches = 0;
		for (int i = 0; i < employees.size(); i++) {
			totalEmployeeFullWatches = totalEmployeeFullWatches + watchesPerPeriod[i];
			totalEmployeePartWatches = totalEmployeePartWatches + partWatchesPerPeriod[i];
		}
		
		if (printOnce) {
			System.out.println("Wymogi obsady kadry");
			
			System.out.println("Konieczna liczba wacht pelnych do obsadzenia kadry w okresie: " + totalEmployeeFullWatches);
			System.out.println("Konieczna liczba wacht resztkowych do obsadzenia kadry w okresie: " + totalEmployeePartWatches);
		}
		
		if (totalEmployeeFullWatches < totalMinFullWatches) {
			System.out.println("Niewystarczajaca liczba kadry do ustalenia obsady.");
			return null;
		}
		if (totalEmployeeFullWatches > totalMaxFullWatches) {
			System.out.println("Zbyt duza obsada kadry w stosunku do wymogow obsady.");
			return null;
		}
		
		for (int i = 0; i < employees.size(); i++) {
			perPersonDayAllocations.add(new ArrayList<IntVar>());
			perPersonNightAllocations.add(new ArrayList<IntVar>());
			perPersonPartAllocations.add(new ArrayList<IntVar>());
		}
		
		List<List<Integer>> empLeaveOfAbsences = new ArrayList<List<Integer>>();
		List<List<Integer>> employeeOffDays = new ArrayList<List<Integer>>();
		List<List<Integer>> employeeOffNights = new ArrayList<List<Integer>>();
		List<List<Integer>> employeeOffParts = new ArrayList<List<Integer>>();
		List<List<Integer>> employeeWorkDays = new ArrayList<List<Integer>>();
		List<List<Integer>> employeeWorkNights = new ArrayList<List<Integer>>();
		List<List<Integer>> employeeWorkParts = new ArrayList<List<Integer>>();
		
		for (int i = 0; i < employees.size(); i++) {
			empLeaveOfAbsences.add(new ArrayList<Integer>());
			employeeOffDays.add(new ArrayList<Integer>());
			employeeOffNights.add(new ArrayList<Integer>());
			employeeOffParts.add(new ArrayList<Integer>());
			employeeWorkDays.add(new ArrayList<Integer>());
			employeeWorkNights.add(new ArrayList<Integer>());
			employeeWorkParts.add(new ArrayList<Integer>());
		}
		
		for (int i = 0; i < employees.size(); i++) {
			for (LeaveOfAbsence loa : employees.get(i).getLeaveOfAbsence()) {
				int dayIndex = calendar.findDayIndex(loa.getDate());
				empLeaveOfAbsences.get(i).add(dayIndex);
				employeeOffDays.get(i).add(dayIndex);
				employeeOffNights.get(i).add(dayIndex);
				employeeOffParts.get(i).add(dayIndex);
			}
			for (ScheduleRequest req : employees.get(i).getAbsenceRequest()) {
				int dayIndex = calendar.findDayIndex(req.getDate());
				if (req.getType().equals("D")) {
					employeeOffDays.get(i).add(dayIndex);
				}
				if (req.getType().equals("N")) {
					employeeOffNights.get(i).add(dayIndex);
				}
				if (req.getType().equals("R")) {
					employeeOffParts.get(i).add(dayIndex);
				}
			}
			for (ScheduleRequest req : employees.get(i).getWorkRequest()) {
				int dayIndex = calendar.findDayIndex(req.getDate());
				if (req.getType().equals("D")) {
					employeeWorkDays.get(i).add(dayIndex);
				}
				if (req.getType().equals("N")) {
					employeeWorkNights.get(i).add(dayIndex);
				}
				if (req.getType().equals("R")) {
					employeeWorkParts.get(i).add(dayIndex);
				}
			}
		}
		
		for (int i = 0; i < calendar.getAllDaysPerPeriod(); i++) {
			//Calendar iDate = calendar.allDays.get(i);
			{
				//String wName = "DAY_" + (i + 1);			
				List<IntVar> perDayFullAllocation = new ArrayList<>();
				for (int j = 0; j < employees.size(); j++) {
					IntVar personDayWatch = buildVariable(definition, employeeOffDays, employeeWorkDays, store, i, j, "D");
					
					perDayFullAllocation.add(personDayWatch);
					perPersonDayAllocations.get(j).add(personDayWatch);
				}
				perDayAllocations.add(perDayFullAllocation);
			}
			
			{
				//String wName = "NIGHT_" + (i + 1);			
				List<IntVar> perNightFullAllocation = new ArrayList<>();
				for (int j = 0; j < employees.size(); j++) {				
					IntVar personNightWatch = buildVariable(definition, employeeOffNights, employeeWorkNights, store, i, j, "N");
					
					perNightFullAllocation.add(personNightWatch);
					perPersonNightAllocations.get(j).add(personNightWatch);
				}
				perNightAllocations.add(perNightFullAllocation);
			}
			
			{
				//String wName = "PART_" + (i + 1);			
				List<IntVar> perPartAllocation = new ArrayList<>();
				for (int j = 0; j < employees.size(); j++) {		
					IntVar personPartWatch = buildVariable(definition, employeeOffParts, employeeWorkParts, store, i, j, "R");
					
					perPartAllocation.add(personPartWatch);
					perPersonPartAllocations.get(j).add(personPartWatch);
				}
				perPartAllocations.add(perPartAllocation);
			}
		}
		
		for (int i = 0; i < calendar.getAllDaysPerPeriod(); i++) {
			allAllocations.addAll(perDayAllocations.get(i));
			allAllocations.addAll(perNightAllocations.get(i));
			allAllocations.addAll(perPartAllocations.get(i));
		}
		
		//--------------------------------------------------------------------------
		//wymogi na dni kaneldarzowe
		for (int i = 0; i < calendar.getAllDaysPerPeriod(); i++) {
			Watch dayDef = daysDefinition.get(calendar.getDayCode(i));
			IntVar[] daySumVars = new IntVar[employees.size()];
			for (int j = 0; j < employees.size(); j++) {
				daySumVars[j] = perDayAllocations.get(i).get(j);
			}
			IntVar daySum = new IntVar(store, "DAYSUM_" + i, dayDef.minDayWatch, dayDef.maxDayWatch);
			store.impose(new SumInt(store, daySumVars, "==", daySum));
			
			IntVar[] nightSumVars = new IntVar[employees.size()];
			for (int j = 0; j < employees.size(); j++) {
				nightSumVars[j] = perNightAllocations.get(i).get(j);
			}
			IntVar nightSum = new IntVar(store, "NIGHTSUM_" + i, dayDef.minNightWatch, dayDef.maxNightWatch);
			store.impose(new SumInt(store, nightSumVars, "==", nightSum));
			
			IntVar[] partSumVars = new IntVar[employees.size()];
			for (int j = 0; j < employees.size(); j++) {
				partSumVars[j] = perPartAllocations.get(i).get(j);
			}
			IntVar partSum = new IntVar(store, "PARTSUM_" + i, dayDef.minPartWatch, dayDef.maxPartWatch);
			store.impose(new SumInt(store, partSumVars, "==", partSum));
			
			if ((dayDef.minDayPartWatch != -1) && (dayDef.maxDayPartWatch != -1)) {
				IntVar[] dayPartSumVars = new IntVar[employees.size() * 2];
				for (int j = 0; j < employees.size(); j++) {
					dayPartSumVars[j] = perDayAllocations.get(i).get(j);
					dayPartSumVars[j + employees.size()] = perPartAllocations.get(i).get(j);
				}
				IntVar dayPartSum = new IntVar(store, "DAYPARTSUM_" + i, dayDef.minDayPartWatch, dayDef.maxDayPartWatch);
				store.impose(new SumInt(store, dayPartSumVars, "==", dayPartSum));
			}
			
			if ((dayDef.min2DayPartWatch != -1) && (dayDef.max2DayPartWatch != -1)) {
				IntVar[] day2PartSumVars = new IntVar[employees.size() * 3];
				for (int j = 0; j < employees.size(); j++) {
					day2PartSumVars[j] = perDayAllocations.get(i).get(j);
					day2PartSumVars[j + employees.size()] = perDayAllocations.get(i).get(j);
					day2PartSumVars[j + employees.size() * 2] = perPartAllocations.get(i).get(j);
				}
				IntVar day2PartSum = new IntVar(store, "DAY2PARTSUM_" + i, dayDef.min2DayPartWatch, dayDef.max2DayPartWatch);
				store.impose(new SumInt(store, day2PartSumVars, "==", day2PartSum));
			}
		}
		
		//--------------------------------------
		//sumaryczna liczba dni pracy w miesiacu
		//--------------------------------------
		int totalMinDays = totalMinFullDayWatches;
		int totalMaxDays = totalMaxFullDayWatches;
		int totalMinNights = totalMinFullNightWatches;
		int totalMaxNights = totalMaxFullNightWatches;
		
		int totalEmployeeMinutes = 0;
		for (int i = 0; i < employees.size(); i++) {
			Employee e = employees.get(i);
			if (e.getNightsRequest() != null) {
				int fixedNights = e.getNightsRequest();
				int fixedDays = watchesPerPeriod[i] - fixedNights;
				
				totalMinDays = totalMinDays - fixedDays;
				totalMaxDays = totalMaxDays - fixedDays;
				totalMinNights = totalMinNights - fixedNights;
				totalMaxNights = totalMaxNights - fixedNights;
			} else {
				totalEmployeeMinutes = totalEmployeeMinutes + minsPerPeriod[i];
			}
		}
		
		if (printOnce) {
			System.out.println("Dniowki do rozdzialu bez okreslonych limitow: " + totalMinDays + " - " + totalMaxDays);
			System.out.println("Nocki do rozdzialu bez okreslonych limitow  : " + totalMinNights + " - " + totalMaxNights);
		}
		
		//sumaryczna liczba dniowek
		for (int i = 0; i < employees.size(); i++) {
			IntVar[] empSumVars = new IntVar[calendar.getAllDaysPerPeriod()];
			for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
				empSumVars[j] = perPersonDayAllocations.get(i).get(j);
			}
			double minutesRatio = minsPerPeriod[i] / (double)totalEmployeeMinutes;
			int minDaily = (int)Math.floor(minutesRatio * totalMinDays);
			int maxDaily = (int)Math.ceil(minutesRatio * totalMaxDays);
			
			if (employees.get(i).getNightsRequest() != null) {
				minDaily = maxDaily = watchesPerPeriod[i] - employees.get(i).getNightsRequest();
			}
			
			if (printOnce) {
				System.out.println("Pracownik[" + (i+1) + "], liczba dyzurow dziennych: " + minDaily + " - " + maxDaily);
			}
			IntVar empSum = new IntVar(store, "DAYEMPSUM_" + i, minDaily, maxDaily);
			store.impose(new SumInt(store, empSumVars, "==", empSum));
		}
		
		//sumaryczna liczba nocek
		for (int i = 0; i < employees.size(); i++) {
			IntVar[] empSumVars = new IntVar[calendar.getAllDaysPerPeriod()];
			for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
				empSumVars[j] = perPersonNightAllocations.get(i).get(j);
			}
			double minutesRatio = minsPerPeriod[i] / (double)totalEmployeeMinutes;
			int minNightly = (int)Math.floor(minutesRatio * totalMinNights);
			int maxNightly = (int)Math.ceil(minutesRatio * totalMaxNights);
			
			if (employees.get(i).getNightsRequest() != null) {
				minNightly = maxNightly = employees.get(i).getNightsRequest();
			}
			
			if (printOnce) {
				System.out.println("Pracownik[" + (i+1) + "], liczba dyzurow nocnych: " + minNightly + " - " + maxNightly);
			}
			IntVar empSum = new IntVar(store, "EMPNIGHTSUM_" + i, minNightly, maxNightly);
			store.impose(new SumInt(store, empSumVars, "==", empSum));
		}
		
		//sumaryczna liczba dniowek i nocek
		for (int i = 0; i < employees.size(); i++) {
			IntVar[] empSumVars = new IntVar[calendar.getAllDaysPerPeriod() * 2];
			for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
				empSumVars[j] = perPersonDayAllocations.get(i).get(j);
			}
			for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
				empSumVars[j + calendar.getAllDaysPerPeriod()] = perPersonNightAllocations.get(i).get(j);
			}
			IntVar empSum = new IntVar(store, "FULLEMPSUM_" + i, watchesPerPeriod[i], watchesPerPeriod[i]);
			store.impose(new SumInt(store, empSumVars, "==", empSum));
		}
		
		//sumaryczna liczba resztek
		for (int i = 0; i < employees.size(); i++) {
			IntVar[] empSumVars = new IntVar[calendar.getAllDaysPerPeriod()];
			for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
				empSumVars[j] = perPersonPartAllocations.get(i).get(j);
			}
			IntVar empSum = new IntVar(store, "EMPPARTSUM_" + i, partWatchesPerPeriod[i], partWatchesPerPeriod[i]);
			store.impose(new SumInt(store, empSumVars, "==", empSum));
		}
		
		//--------------------------------------------
		//w danym dniu tylko dniowka, nocka lub resztka
		for (int i = 0; i < employees.size(); i++) {
			for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
				IntVar day = perPersonDayAllocations.get(i).get(j);
				IntVar night = perPersonNightAllocations.get(i).get(j);
				IntVar part = perPersonPartAllocations.get(i).get(j);
				store.impose(new Not(new XplusYplusQgtC(day, night, part, 1)));
			}
		}
		
		//--------------------------------------
		//dyzury i urlopy
		for (int i = 0; i < employees.size(); i++) {
			List<Integer> leaveOfAbsence = empLeaveOfAbsences.get(i);
			for (int d : leaveOfAbsence) {
				if (d == 0) {
					continue;
				}
				//nie moze byc nocki bezposrednio przed urlopem
				IntVar night = perPersonNightAllocations.get(i).get(d - 1);
				store.impose(new XeqC(night, 0));
			}
		}
		
		//-------------------------------------
		for (int i = 0; i < employees.size(); i++) {
			//nie moze byc dwoch dniowek po sobie
			for (int j = 0; j < calendar.getAllDaysPerPeriod() - 1; j++) {
				IntVar today = perPersonDayAllocations.get(i).get(j);
				IntVar tomorrow = perPersonDayAllocations.get(i).get(j + 1);
				store.impose(new Not(new XplusYgtC(today, tomorrow, 1)));
			}
			//nie moze byc dwoch nocek po sobie
			for (int j = 0; j < calendar.getAllDaysPerPeriod() - 1; j++) {
				IntVar today = perPersonNightAllocations.get(i).get(j);
				IntVar tomorrow = perPersonNightAllocations.get(i).get(j + 1);
				store.impose(new Not(new XplusYgtC(today, tomorrow, 1)));
			}
		}
		
		//-----------------------------------
		for (int i = 0; i < employees.size(); i++) {
			//po nocce nastepna dniowka jest wolna
			for (int j = 0; j < calendar.getAllDaysPerPeriod() - 1; j++) {
				IntVar night = perPersonNightAllocations.get(i).get(j);
				IntVar tomorrowDay = perPersonDayAllocations.get(i).get(j + 1);
				store.impose(new Not(new XplusYgtC(tomorrowDay, night, 1)));
			}
			//po nocce nastepna resztka jest wolna
			for (int j = 0; j < calendar.getAllDaysPerPeriod() - 1; j++) {
				IntVar night = perPersonNightAllocations.get(i).get(j);
				IntVar tomorrowPart = perPersonPartAllocations.get(i).get(j + 1);
				store.impose(new Not(new XplusYgtC(tomorrowPart, night, 1)));
			}
			//po dniowce nastepna resztka jest wolna
			for (int j = 0; j < calendar.getAllDaysPerPeriod() - 1; j++) {
				IntVar day = perPersonDayAllocations.get(i).get(j);
				IntVar tomorrowPart = perPersonPartAllocations.get(i).get(j + 1);
				store.impose(new Not(new XplusYgtC(tomorrowPart, day, 1)));
			}
			//po resztce nastepna dniowka jest wolna
			for (int j = 0; j < calendar.getAllDaysPerPeriod() - 1; j++) {
				IntVar part = perPersonPartAllocations.get(i).get(j);
				IntVar tomorrowDay = perPersonDayAllocations.get(i).get(j + 1);
				store.impose(new Not(new XplusYgtC(part, tomorrowDay, 1)));
			}
		}
		
		//-----------------------------------
		//maksymalna liczba dni pracy w tygodniu
		List<CalendarWeek> weeksDef = null;
		switch (definition.weeklyWorkload) {
		case CALENDAR: weeksDef = calendar.getCalendarWeeks(); break;
		case PERIOD: weeksDef = calendar.getPeriodBasedWeeks(); break;
		case DAILY: weeksDef = calendar.getAllPossibleWeeks(); break;
		}
		for (int w = 0; w < weeksDef.size(); w++) {
			CalendarWeek weekDef = weeksDef.get(w);
			
			for (int i = 0; i < employees.size(); i++) {				
				List<IntVar> weekSumVars = new ArrayList<IntVar>();
				for (int j = weekDef.start; j <= weekDef.end; j++) {
					IntVar day = perPersonDayAllocations.get(i).get(j);
					IntVar night = perPersonNightAllocations.get(i).get(j);
					IntVar part = perPersonPartAllocations.get(i).get(j);
					weekSumVars.add(day);
					weekSumVars.add(night);
					weekSumVars.add(part);
				}
				
				IntVar[] weekSumArray = weekSumVars.toArray(new IntVar[0]);
				
				IntVar empSum = new IntVar(store, "SUMWEEK_" + w + "_EMP_" + i, 0, definition.getWeekWatchLimit());
				store.impose(new SumInt(store, weekSumArray, "==", empSum));
			}			
		}
		
		printOnce = false;
		
		return new ProblemFormulation(store, allAllocations, null);
	}
	
	private ProblemSolution search(ProblemDefinition definition, ProblemFormulation formulation) {
		
		List<CSPVariable> bestAllocation = null;
		int bestScore = 10000;
		
		//SEARCH ROUTINE
        int searchLoop = solverConfig.searchIters;
        while ((searchLoop > 0) && (bestScore > 0)) {
        	System.out.println("Iteracja (" + searchLoop + ")...");
        	searchLoop--;
        	
        	formulation = cspProblemFormulation(definition);
        	
        	Search<IntVar> search = new DepthFirstSearch<IntVar>();
        	search.setTimeOut(solverConfig.searchTime);
        	search.setPrintInfo(false);
        	search.getSolutionListener().searchAll(true);
        	search.getSolutionListener().recordSolutions(true);
        	SelectChoicePoint<IntVar> select = new RandomSelect<IntVar>(formulation.allAllocations.toArray(new IntVar[0]), new IndomainRandom<IntVar>());
        	//SelectChoicePoint<IntVar> select = new InputOrderSelect<IntVar>(formulation.store, formulation.allAllocations.toArray(new IntVar[0]), new IndomainMin<IntVar>());
        	boolean result = search.labeling(formulation.store, select);
        	if (result) {
        		System.out.println("*** Znaleziono rozwiazan: " + search.getSolutionListener().solutionsNo());
        		for (int i = 0; i < search.getSolutionListener().solutionsNo(); i++) {
            		Domain[] solution = search.getSolutionListener().getSolution(i + 1);
            		
            		List<CSPVariable> allocation = new ArrayList<CSPVariable>();
            		for (int j = 0; j < solution.length; j++) {
            			//System.out.println(formulation.allAllocations.get(j).id + " --- " + solution[j].toString());
            			CSPVariable var = new CSPVariable(formulation.allAllocations.get(j).id, solution[j].toString());
            			allocation.add(var);
            		}
            		int score = rankSolution(allocation, definition);
            		if (score < bestScore) {
            			bestScore = score;
            			bestAllocation = allocation;
            			System.out.println("Ocena rozwiazania["+i+"] = " + bestScore);
            		}
            	}
        		
        	} else { 
                System.out.println("*** Brak znalezionych rozwiazan");
            }
        }
        
        if (bestAllocation == null) {
        	return null;
        }
        
        ProblemSolution solution = new ProblemSolution(definition);
    	solution.buildFromSolver6(bestAllocation);
    	solution.solutionScore = bestScore;
        return solution;
	}
	
	public int rankSolution(List<CSPVariable> solution, ProblemDefinition definition) {		
		TimePeriod calendar = definition.getAnalysisPeriod();
		int employees = definition.getEmployees().size();
		
		String [][] data = new String[definition.getEmployees().size()][calendar.getAllDaysPerPeriod()];
		
		for (CSPVariable pAlloc : solution) {
			if (pAlloc.solution == 1) {
				if (data[pAlloc.employee_number][pAlloc.watch_number] != null) {
					System.out.println("ERROR");
					System.exit(0);
				}
				if (pAlloc.dayWatch) {
					data[pAlloc.employee_number][pAlloc.watch_number] = "D";
				}
				if (pAlloc.nightWatch) {
					data[pAlloc.employee_number][pAlloc.watch_number] = "N";
				}
				if (pAlloc.partWatch) {
					data[pAlloc.employee_number][pAlloc.watch_number] = "R";
				}
			}
		}
		
		int score = 0;
		
		if (definition.existingSolution != null) {
			ProblemSolution newSolution = new ProblemSolution(definition);
			newSolution.buildFromSolver6(solution);
			
			for (int i = 0; i < employees; i++) {
				int penalty = 1;
				for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
					String newValue = newSolution.data[i][j];
					String oldValue = definition.existingSolution.data[i][j];
					if (!Objects.equals(newValue, oldValue)) {
						score = score + penalty;
						penalty++;
					}
				}
			}
		}
		
		int totalHolidayWatches = 0;
		for (int i = 0; i < employees; i++) {
			for (int j = 0; j < calendar.dayCodes.length; j++) {
				if (calendar.dayCodes[j].contains("*") && (data[i][j] != null)) {
					if (data[i][j].equals("D") || data[i][j].equals("N") || data[i][j].equals("R")) {
						totalHolidayWatches = totalHolidayWatches + 1;
					}
				}
			}
		}
		if (totalHolidayWatches > 0) {
			//liczmy tylko wtedy, gdy mamy w kalendarzu zdefiniowane dni swiateczne
			int totalMins = 0;
			for (int i = 0; i < employees; i++) {
				totalMins = totalMins + definition.employeeMinsPerPeriod(i);
			}
			for (int i = 0; i < employees; i++) {
				int mins = definition.employeeMinsPerPeriod(i);
				int holidayWatches = 0;
				for (int j = 0; j < calendar.dayCodes.length; j++) {
					if (calendar.dayCodes[j].contains("*") && (data[i][j] != null)) {
						if (data[i][j].equals("D") || data[i][j].equals("N") || data[i][j].equals("R")) {
							holidayWatches = holidayWatches + 1;
						}
					}
				}
				double minsRatio = mins / (double)totalMins;
				//double holidayRatio = holidayWatches / (double)totalHolidayWatches;
				//System.out.println("RATIO: " + minsRatio + " - " + (minsRatio * totalHolidayWatches) + " - " + holidayRatio + " - " + holidayWatches + " - " + totalHolidayWatches);
				
				double x = minsRatio * totalHolidayWatches;
				double y = holidayWatches;
				
				double d = (x - y) * (x - y);
				score = score + (int)(10 * d);
				//System.out.println("EMP: " + i + " D: " + d);
			}
		}
		
		for (int i = 0; i < employees; i++) {
			int penalty = 1;
			for (int j = 0; j < calendar.getAllDaysPerPeriod() - 1; j++) {
				if ((data[i][j] != null) && (data[i][j + 1] != null)) {
					score = score + penalty;
					penalty++;
				}
			}
		}
		return score;
	}
}
