/*
  This file is part of Solver.

    Solver is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    Solver is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Solver. If not, see <http://www.gnu.org/licenses/>.
*/

package infra.definition;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.Map;

public class TimePeriod {
	public Calendar startDate;
	public Calendar endDate;
	
	private String[] weekCodes;
	
	public List<Calendar> allDays;
	public Map<Calendar, String> specialDays;
	
	public int[] dayOfWeek;
	public String[] dayCodes;
	public boolean[] workingDays;
	
	public TimePeriod(Calendar startDate, Calendar endDate, String[] weekCodes, Map<Calendar, String> specialDays) {
		this.startDate = startDate;
		this.endDate = endDate;
		this.weekCodes = weekCodes;
		this.specialDays = specialDays;
		
		this.allDays = new ArrayList<Calendar>();
		Calendar c = (Calendar)this.startDate.clone();
		while (!c.after(this.endDate)) {
			this.allDays.add(c);
			c = (Calendar)c.clone();
			c.add(Calendar.DATE, 1);
		}

		dayOfWeek = new int[this.allDays.size()];
		workingDays = new boolean[this.allDays.size()];
		dayCodes = new String[this.allDays.size()];
		
		for (int i = 0; i < this.allDays.size(); i++) {
			Calendar day = this.allDays.get(i);
			dayOfWeek[i] = day.get(Calendar.DAY_OF_WEEK);
			dayCodes[i] = this.weekCodes[dayOfWeekToIndex(dayOfWeek[i])];
			if (specialDays.containsKey(day)) {
				dayCodes[i] = specialDays.get(day);
			}
			
			workingDays[i] = !dayCodes[i].contains("*");
		}
	}
	
	private int dayOfWeekToIndex(int dayOfWeek) {
		if (dayOfWeek == 1) {
			return 6;
		}
		return dayOfWeek - 2;
	}
	
	public int getAllDaysPerPeriod() {
		return this.allDays.size();
	}
	
	public int getWorkDaysPerPeriod() {
		int workDaysPerPeriod = 0;
		for (int i = 0; i < this.allDays.size(); i++) {
			if (workingDays[i]) {
				workDaysPerPeriod++;
			}
		}
		return workDaysPerPeriod;
	}
	
	public int findDayIndex(Calendar cal) {
		for (int i = 0; i < this.allDays.size(); i++) {
			Calendar day = this.allDays.get(i);
			if (day.equals(cal)) {
				return i;
			}
		}
		return -1;
	}
	
	public int getWorkDaysPerPeriod(List<LeaveOfAbsence> leaveOfAbsence) {
		boolean[] workArray = new boolean[this.allDays.size()];
		for (int i = 0; i < workArray.length; i++) {
			workArray[i] = workingDays[i];
		}
		for (LeaveOfAbsence ansence : leaveOfAbsence) {
			for (int i = 0; i < this.allDays.size(); i++) {
				Calendar day = this.allDays.get(i);
				if (day.equals(ansence.getDate())) {
					workArray[i] = false;
				}
			}
		}

		int totalWorkDays = 0;
		for (int i = 0; i < workArray.length; i++) {
			if (workArray[i]) {
				totalWorkDays++;
			}
		}
		return totalWorkDays;
	}
	
	public int getMinutesPerPeriod(Employee emp) {
		double minsPerDay = emp.getMinsPerDay();
		
		double[] minArray = new double[this.allDays.size()];
		for (int i = 0; i < minArray.length; i++) {
			minArray[i] = workingDays[i] ? minsPerDay : 0;
		}
		for (LeaveOfAbsence absence : emp.getLeaveOfAbsence()) {
			for (int i = 0; i < this.allDays.size(); i++) {
				Calendar day = this.allDays.get(i);
				if (day.equals(absence.getDate())) {
					if (absence.getAbsenceTime() == 0) {
						minArray[i] = 0;
					} else {
						//THIS VALUE CAN BE NEGATIVE (you can deduct more time per day)
						minArray[i] = minsPerDay - absence.getAbsenceTime();
					}
				}
			}
		}
		
		double totalWorkMins = 0;
		for (int i = 0; i < minArray.length; i++) {
			totalWorkMins = totalWorkMins + minArray[i];
		}
		return (int)totalWorkMins;
	}
	
	/*@Deprecated
	public List<Integer> getOffDaysPerPeriod(List<Calendar> leaveOfAbsence, List<Calendar> offDays) {
		List<Integer> indexes = new ArrayList<Integer>();
		for (int i = 0; i < this.allDays.size(); i++) {
			Calendar day = this.allDays.get(i);
			for (Calendar absence : leaveOfAbsence) {
				if (day.equals(absence)) {
					indexes.add(i);
				}
			}
			for (Calendar offDay : offDays) {
				if (day.equals(offDay)) {
					indexes.add(i);
				}
			}
		}
		return indexes;
	}*/
		
	public boolean isNonWorkingDay(int dayIndex) {
		return !workingDays[dayIndex];
	}
	
	public String getDayOfWeek(int dayIndex) {
		switch (dayOfWeek[dayIndex]) {
			case Calendar.MONDAY: return "Po";
			case Calendar.TUESDAY: return "Wt";
			case Calendar.WEDNESDAY: return "Sr";
			case Calendar.THURSDAY: return "Cz";
			case Calendar.FRIDAY: return "Pt";
			case Calendar.SATURDAY: return "So";
			case Calendar.SUNDAY: return "N";
		}
		return "";
	}
		
	public String getDayCode(int dayIndex) {
		return dayCodes[dayIndex];
	}
	
	public List<CalendarWeek> getPeriodBasedWeeks() {
		List<CalendarWeek> result = new ArrayList<>();
		for (int day = 0; day < getAllDaysPerPeriod(); day = day + 7) {
			int start = day;
			int end = Math.min(start + 6, getAllDaysPerPeriod() - 1);
			CalendarWeek week = new CalendarWeek(start, end);
			result.add(week);
		}
		return result;
	}
	
	public List<CalendarWeek> getCalendarWeeks() {
		List<CalendarWeek> result = new ArrayList<>();
		int start = 0;
		for (int day = 0; day < getAllDaysPerPeriod(); day++) {
			if (dayOfWeek[day] == Calendar.SUNDAY) {
				CalendarWeek week = new CalendarWeek(start, day);
				result.add(week);
				start = day + 1;
			}
		}
		if (start < getAllDaysPerPeriod()) {
			CalendarWeek week = new CalendarWeek(start, getAllDaysPerPeriod() - 1);
			result.add(week);
		}
		return result;
	}
	
	public List<CalendarWeek> getAllPossibleWeeks() {
		List<CalendarWeek> result = new ArrayList<>();
		for (int day = 0; day < getAllDaysPerPeriod(); day++) {
			int start = day;
			int end = Math.min(start + 6, getAllDaysPerPeriod() - 1);
			CalendarWeek week = new CalendarWeek(start, end);
			result.add(week);
		}
		return result;
	}
}
