/*
  This file is part of Solver.

    Solver is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    Solver is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Solver. If not, see <http://www.gnu.org/licenses/>.
*/

package infra.solution;

import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import infra.Utils;
import infra.definition.Employee;
import infra.definition.LeaveOfAbsence;
import infra.definition.ProblemDefinition;
import infra.definition.TimePeriod;
import infra.definition.Watch;
import infra.formulation.CSPVariable;

public class ProblemSolution {
	
	public String[][] data;
	private ProblemDefinition definition;
	
	public int solutionScore;
	
	public ProblemSolution(ProblemDefinition definition) {
		this.definition = definition;
	}
	
	public void buildFromSolver6(List<CSPVariable> solution) {
		TimePeriod calendar = definition.getAnalysisPeriod();
		
		data = new String[definition.getEmployees().size()][calendar.getAllDaysPerPeriod()];
		for (int i = 0; i < definition.getEmployees().size(); i++) {
			Employee employee = definition.getEmployees().get(i);
			for (LeaveOfAbsence loa : employee.getLeaveOfAbsence()) {
				int j = calendar.findDayIndex(loa.getDate());
				if (j < 0) {
					System.err.println("Brak zgodnosci kalendarza w definicji problemu i rozwiazaniu - dni wolne.");
					System.exit(0);
				}
				data[i][j] = "U";
			}
			
			/*for (int j : calendar.getOffDaysPerPeriod(employee.getLeaveOfAbsence(), new ArrayList<Calendar>())) {
				data[i][j] = "U";
			}*/
			/*for (int j : calendar.getOffDaysPerPeriod(new ArrayList<Calendar>(), employee.getAbsenceRequest())) {
				data[i][j] = "W";
			}*/
		}
		
		for (CSPVariable pAlloc : solution) {
			if (pAlloc.solution == 1) {
				if (data[pAlloc.employee_number][pAlloc.watch_number] != null) {
					System.out.println("ERROR");
					System.exit(0);
				}
				if (pAlloc.dayWatch) {
					data[pAlloc.employee_number][pAlloc.watch_number] = "D";
				}
				if (pAlloc.nightWatch) {
					data[pAlloc.employee_number][pAlloc.watch_number] = "N";
				}
				if (pAlloc.partWatch) {
					data[pAlloc.employee_number][pAlloc.watch_number] = "R";
				}
			}
		}
	}
	
	public boolean buildFromFile(String fileName) {
		
		TimePeriod calendar = definition.getAnalysisPeriod();
		List<Employee> employees = definition.getEmployees();
		data = new String[employees.size()][calendar.getAllDaysPerPeriod()];
		
		try {
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document doc = builder.parse(fileName);
		
			XPathFactory xPathfactory = XPathFactory.newInstance();
			XPath xpath = xPathfactory.newXPath();
			
			NodeList employeeNodes = (NodeList)xpath.compile("solution/employee").evaluate(doc, XPathConstants.NODESET);
			for (int i = 0; i < employeeNodes.getLength(); i++) {
				String employeeCode = xpath.compile("./@code").evaluate(employeeNodes.item(i));
				int index = -1;
				for (int e = 0; e < employees.size(); e++) {
					if (employees.get(e).getCode().equals(employeeCode)) {
						index = e;
					}
				}
				if (index == -1) {
					System.err.println("Brak zgodnosci kodow pracownikow w definicji problemu i rozwiazaniu.");
					return false;
				}
				
				NodeList eventNodes = (NodeList)xpath.compile("./event").evaluate(employeeNodes.item(i), XPathConstants.NODESET);
				for (int j = 0; j < eventNodes.getLength(); j++) {
					String eventDate = xpath.compile("./@date").evaluate(eventNodes.item(j));
					String eventType = xpath.compile("./@type").evaluate(eventNodes.item(j));
					
					Calendar date = Utils.parse(eventDate);
					int dayIndex = calendar.findDayIndex(date);
					if (dayIndex < 0) {
						System.err.println("Brak zgodnosci kalendarza w definicji problemu i rozwiazaniu.");
						return false;
					}
					
					data[index][dayIndex] = eventType;
				}
			}
			
		} catch (Exception ex) {
			ex.printStackTrace();
			return false;
		}
		return true;
	}
	
	public boolean serializeToXml(String fileName) {
		
		DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
        DocumentBuilder dBuilder;
        
        try {
        	dBuilder = dbFactory.newDocumentBuilder();
        	Document doc = dBuilder.newDocument();
        	Element rootElement = doc.createElement("solution");
        	doc.appendChild(rootElement);
        	
        	TimePeriod calendar = definition.getAnalysisPeriod();
        	List<Employee> employees = definition.getEmployees();
        	for (int i = 0; i < employees.size(); i++) {
        		Employee employee = employees.get(i);
        		Element xmlEmployee = doc.createElement("employee");
        		xmlEmployee.setAttribute("code", employee.getCode());
        		
        		for (int j = 0; j < data[i].length; j++) {
        			if (data[i][j] != null) {
        				Element xmlEvent = doc.createElement("event");
        				Calendar day = calendar.allDays.get(j);
        				xmlEvent.setAttribute("type", data[i][j]);
        				xmlEvent.setAttribute("date", Utils.build(day));
        				xmlEmployee.appendChild(xmlEvent);
        			}
        		}
        		
        		rootElement.appendChild(xmlEmployee);
        	}
        	
        	Element metadata = doc.createElement("metadata");
        	Element metadataScore = doc.createElement("score");
        	metadataScore.setTextContent("" + solutionScore);
        	metadata.appendChild(metadataScore);
        	rootElement.appendChild(metadata);
				
        	DOMSource source = new DOMSource(doc);
        	StreamResult file = new StreamResult(fileName);
        	TransformerFactory transformerFactory = TransformerFactory.newInstance();
        	Transformer transformer = transformerFactory.newTransformer();
        	transformer.setOutputProperty(OutputKeys.INDENT, "yes");
        	transformer.transform(source, file);
        } catch (Exception ex) {
        	ex.printStackTrace();
        }
		
		return true;
	}
	
	public void printout() {
		
		TimePeriod calendar = definition.getAnalysisPeriod();
		
		System.out.println("-------------");
		SimpleDateFormat fmt = new SimpleDateFormat("dd-MMM-yyyy");
		System.out.println("Data poczatku okresu  : " + fmt.format(calendar.startDate.getTime()));
		System.out.println("Data konca okresu     : " + fmt.format(calendar.endDate.getTime()));
		System.out.println("-------------");
		
		System.out.print("DZ  |");
		for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
			System.out.format("%2d", (j + 1));
			System.out.print("|");
		}
		System.out.println();
		System.out.print("    |");
		for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
			System.out.format("%2s", calendar.getDayOfWeek(j));
			System.out.print("|");
		}
		System.out.println();
		System.out.print("    |");
		for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
			String dayCode = calendar.getDayCode(j);
			System.out.format("%2s|", dayCode);
		}
		System.out.println();
		System.out.print("----+");
		for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
			System.out.print("--+");
		}
		System.out.println();
		
		Map<String, int[]> counter = new HashMap<String, int[]>();
		//List<CSPWeek> weeksDef = calendar.getWeeks();
		for (int i = 0; i < definition.getEmployees().size(); i++) {
			Employee e = definition.getEmployees().get(i);
			System.out.format("%4s|", e.getCode());
			Map<String, Integer> ppCounter = new HashMap<String, Integer>();
			
			int holidays = 0;
			for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
				boolean isHoliday = calendar.dayCodes[j].contains("*");
				
				String toPrint = data[i][j] != null ? data[i][j] : "";
				if (data[i][j] != null) {
					if (!counter.containsKey(data[i][j])) {
						counter.put(data[i][j], new int[calendar.getAllDaysPerPeriod()]);
					}
					counter.get(data[i][j])[j]++;
					if (!ppCounter.containsKey(data[i][j])) {
						ppCounter.put(data[i][j], 0);
					}
					ppCounter.put(data[i][j], ppCounter.get(data[i][j]) + 1);
					if ((isHoliday) && (data[i][j].equals("D") || data[i][j].equals("N") || data[i][j].equals("R"))) {
						holidays++;
					}
				}
				System.out.format("%2s", toPrint);
				System.out.print("|");
			}
			System.out.format("%3d", holidays);
			System.out.print("**|");
			
			String[] keys = ppCounter.keySet().toArray(new String[0]);
			Arrays.sort(keys);
			for (String key : keys) {
				System.out.format("%3d", ppCounter.get(key));
				System.out.print(" " + key + "|");
			}
			System.out.println("");
			
			System.out.print("----+");
			for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
				System.out.print("--+");
			}
			System.out.println();
		}
		
		String[] keys = counter.keySet().toArray(new String[0]);
		Arrays.sort(keys);
		for (String key : keys) {
			System.out.print("   " + key + "|");
			for (int j = 0; j < calendar.getAllDaysPerPeriod(); j++) {
				System.out.format("%2d|", counter.get(key)[j]);
			}
			System.out.println();
		}
		
		System.out.println("Legenda:");
		System.out.println("PPx = plan pracownika 'x'");
		for (Watch w : definition.getDaysDefinition().values()) {
			System.out.format("%2s  = %s\n", w.code, w.name);
		}
		
		//System.out.println(" *  = dzien wolny");
		//System.out.println(" Z  = dzien zabiegowy");
		
		System.out.println(" U  = urlop");
		//System.out.println(" W  = dzien wolny od pracy (nie urlop)");
		System.out.println(" D  = pelna zmiana dzienna");
		System.out.println(" N  = pelna zmiana nocna");
		System.out.println(" R  = zmiana resztkowa (dzienna)");
	}

}
